(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_task_server.ml 282 2010-10-28 00:13:50Z gerd $ *)

open Printf

type sh_state =
    { pid_of_req_id : (int, int) Hashtbl.t }

module State_shvar =
  Netplex_sharedvar.Make_var_type(struct type t = sh_state end)

let state_name = "mapred_task_server.state"
let mutex_name = state_name ^ ".mutex"

let m = ref None

let change_state op =
  let mutex = Netplex_mutex.access mutex_name in
  Netplex_mutex.lock mutex;
  try
    ( match op with
	| `Reg_req(req_id,pid) ->
	    let st = State_shvar.get state_name in
	    Hashtbl.replace st.pid_of_req_id req_id pid;
	    State_shvar.set state_name st
	| `Del_req req_id ->
	    let st = State_shvar.get state_name in
	    Hashtbl.remove st.pid_of_req_id req_id;
	    State_shvar.set state_name st
	| `Init ->
	    let st = 
	      { pid_of_req_id = Hashtbl.create 5 } in
	    let ok =
	      Netplex_sharedvar.create_var ~enc:true state_name in
	    if ok then
	      State_shvar.set state_name st
    );
    Netplex_mutex.unlock mutex;
  with
    | error ->
	Netplex_mutex.unlock mutex;
	raise error


let redirect_log tm prefix f arg =
  let log_dir = tm#log_directory in
  ( try Unix.mkdir log_dir 0o777
    with Unix.Unix_error (Unix.EEXIST,_,_) -> ()
  );
  let levels = [ `Debug; `Info; `Warning; `Crit ] in
  let files =
    List.map
      (fun level ->
	 let level_s = Netlog.string_of_level level in
	 let name = sprintf "%s/%s%s.log" log_dir prefix level_s in
	 open_out name
      )
      levels in
  let logger lev msg =
    List.iter2
      (fun level ch ->
	 Netlog.channel_logger ch level lev msg
      )
      levels
      files in
  let new_stderr_name = sprintf "%s/%sstderr.log" log_dir prefix in
  let new_stderr = 
    Unix.openfile 
      new_stderr_name
      [Unix.O_CREAT; Unix.O_WRONLY; Unix.O_APPEND]
      0o666 in
  let old_logger = !Netlog.current_logger in
  let old_stderr = Unix.dup Unix.stderr in
  let revert () =
    Netlog.current_logger := old_logger;
    Unix.dup2 old_stderr Unix.stderr;
    Unix.close old_stderr;
    List.iter (fun ch -> close_out ch) files  in
  Netlog.current_logger := logger;
  Unix.dup2 new_stderr Unix.stderr;
  Unix.close new_stderr;
  try
    let r = f arg in
    revert();
    r
  with
    | error ->
	Netlog.logf `Err "Exception: %s" (Netexn.to_string error);
	revert();
	raise error


let upload_log me tm prefix filter_flag =
  try
    let files = tm # task_files_log_directory() in
    let prefix_l = String.length prefix in
    let files' =
      if filter_flag then
	List.filter
	  (fun name ->
	     String.length name >= prefix_l &&
	       String.sub name 0 prefix_l = prefix
	  )
	  files
      else
	files  in
    let c = me # open_cluster (Unixqueue.create_unix_event_system()) in
    try
      tm # task_save_log_directory c files';
      Plasma_client.abort_cluster c
    with error -> Plasma_client.abort_cluster c; raise error
  with
    | error ->
	Netlog.logf `Err "Exception upload_log: %s" (Netexn.to_string error)	


(* needed config:
   - mapred_env:
     * cluster name, namenodes
     * config file
     * arguments (string array) --> exec!
 *)

let proc_exec (mj,conf,cf) (req_id, jobconfig_s, enc_task) =
  let shm_mng = new Plasma_shm.netplex_shm_manager() in
  let me =
    ( object
	method open_cluster esys =
	  let c =
	    Plasma_client.open_cluster 
	      conf#nn_clustername
	      (List.map Plasma_util.parse_host_port conf#nn_nodes)
	      esys in
	  Plasma_client.configure_shm_manager c shm_mng;
	  c
	    
	method config = conf
	method config_file = cf
      end
    ) in
  let tm_opt = ref None in
  let task_prefix = sprintf "task_%06d_" req_id in
  try
    change_state (`Reg_req(req_id, Unix.getpid()));
    let mjc = Mapred_job_config.unmarshal jobconfig_s in
    let jc = Mapred_job_config.mapred_job_config mjc in
    let task = Mapred_tasks.decode_task enc_task in
    let tm = Mapred_taskfiles.scp_taskfile_manager me#config jc in
    tm_opt := Some tm;
    let ti =
      ( object
	  method req_id = req_id
	  method task_prefix = task_prefix
	  method partition = 
	    match task with
	      | `Shuffle s when s.Mapred_tasks.shuffle_reduce -> 
		  fst s.Mapred_tasks.shuffle_partitions
	      | _ ->
		  failwith "partition: not in a reduce task"
	end
      ) in
    let r = 
      redirect_log tm task_prefix
	(Mapred_task_exec.execute me jc mj task) ti in
    upload_log me tm task_prefix (task <> `Cleanup);
    if task = `Cleanup then
      tm # task_delete_log_directory();
    change_state (`Del_req req_id);
    Mapred_tasks.encode_task_result r
  with
    | error ->
	Netlog.logf `Err "Exception in exec (pid %d): %s"
	  (Unix.getpid()) (Netexn.to_string error);
	( match !tm_opt with
	    | None -> ()
	    | Some tm ->
		upload_log me tm task_prefix true
	);
	Mapred_tasks.encode_task_result
	  (`Error (Netexn.to_string error))


let proc_kill req_id =
  let st = State_shvar.get state_name in
  try
    let pid = Hashtbl.find st.pid_of_req_id req_id in
    ( try
	Unix.kill pid Sys.sigterm
      with
	| error ->
	    Netlog.logf `Err "Cannot kill pid %d: %s"
	      pid
	      (Netexn.to_string error);
    );
    change_state (`Del_req req_id)
  with
    | Not_found ->
	()


let proc_kill_all () =
  let st = State_shvar.get state_name in
  Hashtbl.iter
    (fun req_id pid ->
       ( try
	   Unix.kill pid Sys.sigterm
	 with
	   | error ->
	       Netlog.logf `Err "Cannot kill pid %d: %s"
		 pid
		 (Netexn.to_string error);
       );
       change_state (`Del_req req_id)
    )
    st.pid_of_req_id


let proc_configure (low,high) =
  Plasma_shm.Netplex_shm_manager.init ~low ~high


let real_version =
  Mapred_config.executable_version()

let proc_check_version expected_version =
  expected_version = real_version


let setup srv (mj,conf,cf) =
  Mapred_rpc_srv.Mapred_task.V1.bind
    ~proc_null:(fun () -> ())
    ~proc_exec:(proc_exec (mj,conf,cf))
    ~proc_kill
    ~proc_kill_all
    ~proc_configure
    ~proc_check_version
    srv;
  Rpc_server.set_exception_handler
    srv
    (fun e ->
       prerr_endline(Netexn.to_string e)
    )


let factory mj job_cf =
  Rpc_netplex.rpc_factory
   ~configure:(fun _ _ -> 
		 (mj, Mapred_config.extract_config job_cf, job_cf))
    ~hooks:(fun _ ->
              ( object
                  inherit Netplex_kit.empty_processor_hooks()
                    
                  method post_add_hook socksrv ctrl =
		    ctrl # add_plugin Netplex_mutex.plugin;
		    ctrl # add_plugin Netplex_sharedvar.plugin		    

                  method post_start_hook cont =
		    change_state `Init;

		    Rpc_server.Debug.enable := true;
		    Rpc_server.Debug.enable_ctrace := true;
		    Rpc_server.Debug.enable_ptrace := true;
		end
	      )
	   )
    ~setup
    ~name:"mapred_task_server"
    ()
